#!/usr/bin/env python

import ase
import ase.geometry
import numpy as np
from ase.calculators.calculator import Calculator
from ase.constraints import FixAtoms
from ase.optimize import BFGS


def calc_aggregated_uncertainties(
    atom_pos: np.ndarray,
    unit_cell: np.ndarray,
    atomic_uncertainties: np.ndarray,
    boundary_conditions: np.ndarray = np.ones(3, dtype=bool),
    r_cut: float = 4.0,
) -> np.ndarray:
    """Calculate the locally aggregated uncertainties for a single structure

    Args:
        atom_pos: Atomic positions, shape=(n_atoms, 3)
        unit_cell: Unit cell, shape=(3, 3)
        atomic_uncertainties: Atomic force uncertainties, shape=(n_atoms, 3)
        boundary_conditions: Array of booleans indicating boundary conditions,
            shape=(3)
        r_cut: Cutoff radius defining the aggregation region

    Returns:
        Aggregated uncertainties for each atom, shape=(n_atoms)
    """
    radii = ase.geometry.get_distances(
        atom_pos, cell=unit_cell, pbc=boundary_conditions
    )[1]
    mask = radii < r_cut
    unc_mean = atomic_uncertainties.mean(axis=-1)
    return (mask * unc_mean[np.newaxis, :]).sum(axis=1) / mask.sum(axis=1)


def cut_and_relax(
    atom_pos: np.ndarray,
    atom_types: np.ndarray,
    unit_cell: np.ndarray,
    center_idx: int,
    calculator: Calculator,
    molecules_in_box: int = 64,
    freeze_fraction: float = 0.8,
    max_vacuum: float = 2.0,
    step_size: float = 0.2,
) -> ase.Atoms:
    """Cut out a smaller box out of a single frame. The position and types
    arrays are assumed to be ordered in OH2 units ['O', 'H', 'H'] * n_molecules

    Args:
        atom_pos: Atomic positions, shape=(n_atoms, 3)
        atom_types: Atom types, shape=(n_atoms)
        unit_cell: Matrix representation of unit cell, shape=(3, 3)
        center_idx: Index of the atom in the center of the box.
            The environment around this atom is fixed during relaxation
        calculator: ASE calculator to use for relaxation
        molecules_in_box: Total number of molecules in the cut-out box
        freeze_fraction: Volume fraction of the cut box to fix atoms in during
            relaxation
        max_vacuum: Length of vacuum in each direction for the first
            minimization iteration
        step_size: Annealing step size

    Returns:
        ASE atoms representation of the cut-out box
    """
    n_atoms = atom_types.shape[0]
    dists, radii = ase.geometry.get_distances(
        atom_pos, cell=unit_cell, pbc=np.ones(3, dtype=bool)
    )
    oxygen_idx = np.arange(n_atoms)[atom_types == 8][
        np.argsort(
            np.max(np.fabs(dists[center_idx][atom_types == 8]), axis=1)
        )[:molecules_in_box]
    ]
    hydrogen_idx = np.arange(n_atoms)[atom_types == 1][
        np.argsort(np.min(radii[oxygen_idx], axis=0)[atom_types == 1])[
            : 2 * molecules_in_box
        ]
    ]
    all_atom_idx = np.sort(np.concatenate((oxygen_idx, hydrogen_idx)))
    box_length = (
        np.prod(np.diag(unit_cell)) * len(oxygen_idx) / (len(atom_pos) / 3)
    ) ** (1.0 / 3.0)
    sel_positions = dists[center_idx, all_atom_idx] + np.array(
        [box_length / 2] * 3
    )
    sel_types = atom_types[all_atom_idx]

    cut_box = ase.Atoms(
        numbers=sel_types,
        positions=sel_positions,
        cell=[box_length] * 3,
        pbc=np.ones(3, dtype=bool),
    )
    cut_box.calc = calculator
    fix_constraint = FixAtoms(
        mask=np.max(np.fabs(dists[center_idx][all_atom_idx]), axis=1)
        < freeze_fraction * box_length / 2
    )
    cut_box.set_constraint(fix_constraint)

    for vac in np.linspace(
        start=max_vacuum, stop=0.0, num=int(max_vacuum / step_size) + 1
    ):
        cut_box.set_cell([box_length + vac] * 3, scale_atoms=False)
        minimizer = BFGS(cut_box)
        minimizer.run(fmax=0.1, steps=5)
    return cut_box
